#ifndef FI_TREE_CONVERTER_H
#define FI_TREE_CONVERTER_H
//---------------------------------------------------------------------------
//؍\؍\ւ̕ϊ(FIHalfTreeConverter)ƁA
//̕ϊ[Lq邽߂̊ep^[(FIPattern)B
//---------------------------------------------------------------------------
#include <utility>
#include <map>
#include <list>
#include <vector>
#include <boost/shared_ptr.hpp>
#include <boost/spirit/core.hpp>
#include <boost/spirit/tree/parse_tree.hpp>
//---------------------------------------------------------------------------
//xtm[h̎
enum FILabeledType{
	filtNodes,//XȂ}b`ȎΏۂɂȂm[h
	filtValue//ȏ}b`OAlƂĈ
};

//---------------------------------------------------------------------------
//؂̕ϊŕKvȁA^̑gݍ킹ƂɓƎ̒`Kvȑ̏W
//iꉻĎgj
template <class TL, class TR>
struct FIConvertationTraits;
//---------------------------------------------------------------------------
//ؕϊñNXŋʂɎg`WipĎgj
template <class TTraits>
class FITreeDefs
{
public:
	//̃L[̌^B
	typedef typename TTraits::TKey TKey;
	//̌^B
	typedef typename TTraits::TString TString;

  //m[ȟ^i͑jB
	typedef typename TTraits::TInNode TInNode;
  //ľ^i͑jB
	typedef typename TTraits::TInValue TInValue;
  //̌^i͑jB
	typedef typename TTraits::TInAttribute TInAttribute;
	//m[h̊er߂Ce[^B
	typedef typename TTraits::TInIterator TInIterator;

	//̃L[B
	static TKey GetKey(const TInAttribute& attr){
		return TTraits::GetKey(attr);
	}
  //̒lB
	static TInValue GetValue(const TInAttribute& attr){
		return TTraits::GetValue(attr);
	}
  //lȂtrueAm[hȂfalseB
	static bool IsString(const TInValue& value){
		return TTraits::IsString(value);
	}
	//l̕B
	static TString GetString(const TInValue& value){
		return TTraits::GetString(value);
	}
  //l̃m[hB
	static TInNode GetNode(const TInValue& value){
		return TTraits::GetNode(value);
	}
	//qm[h̐擪擾
	static TInIterator GetBegin(const TInNode& node){
		return TTraits::GetBegin(node);
	}
	//qm[h̖̎擾
	static TInIterator GetEnd(const TInNode& node){
		return TTraits::GetEnd(node);
	}
  //iwB
	static TInAttribute GetAttribute(const TInNode& node, TInIterator i){
		return TTraits::GetAttribute(node, i);
	}

  //m[ȟ^io͑jB
	typedef typename TTraits::TOutNode TOutNode;
  //ľ^io͑jB
  typedef typename TTraits::TOutValue TOutValue;
	//̌^io͑jB
	typedef typename TTraits::TOutAttribute TOutAttribute;
	//̌^io͑jB
	typedef typename TTraits::TOutAttributes TOutAttributes;

  //m[hB
	static void CreateNode(TOutNode& node, const TOutAttributes& attrs){
		TTraits::CreateNode(node, attrs);
	}
  //lB
  static void CreateValue(TOutValue& value, const TString& s){
		TTraits::CreateValue(value, s);
	}
  //m[hlB
  static void CreateValue(TOutValue& value, const TOutNode& node){
		TTraits::CreateValue(value, node);
	}
  //B
  static void CreateAttribute(TOutAttribute& attr, const TKey& key, const TOutValue& value){
		TTraits::CreateAttribute(attr, key, value);
	}
  //̖ɑǉB
  static void Add(TOutAttributes& attrs, const TOutAttribute& attr){
		TTraits::Add(attrs, attr);
	}
  //̖ɑǉB
  static void Add(TOutAttributes& attrs, const TOutAttributes& added){
		TTraits::Add(attrs, added);
	}

	//xtꂽm[h̏
	struct TLabeledInfo{
		TInNode node;//m[h
		TInIterator first;//m[h̐擪
		TInIterator last;//m[h̖̎
		TLabeledInfo(){}
		TLabeledInfo(const TInNode& n, TInIterator f, TInIterator l)
			: node(n), first(f), last(l){}
	};
	//l̔zio͑jB
	typedef std::vector<TOutValue> TOutValues;

	//m[h}b`OɁAxɑΉlL^邽߂̃}bvB
	typedef std::map<int, TLabeledInfo> TSourceLabelMap;
	struct TSourceLabelMaps{
		TSourceLabelMap values;
		TSourceLabelMap attrs;
	};
	//m[h񐶐ɁAxɑΉl^邽߂̃}bvB
	typedef std::map<int, TOutValues> TTargetValueLabelMap;
	typedef std::map<int, TOutAttributes> TTargetAttributeLabelMap;
	struct TTargetLabelMaps{
		TTargetValueLabelMap values;
		TTargetAttributeLabelMap attrs;
	};

};
//---------------------------------------------------------------------------
//؂̃p^[̎̊NX
template <class TTraits>
class FIPatternImpl: public FITreeDefs<TTraits>
{
public:
	FIPatternImpl(): label(0){}
	virtual ~FIPatternImpl(){}
	//m[h񂪃p^[Ƀ}b`邩𒲂ׂ
	virtual std::pair<bool, TInIterator>
		Match(const TInNode& pnode, TInIterator first, TInIterator last, TSourceLabelMaps* lmap)= 0;
	//p^[𐗌^ƂđAattrsɒǉB
	//ɃxtĂAxɑΉ鑮ԂB
	void Generate(TOutAttributes& attrs, const TTargetLabelMaps* lmap);
	//xt
	void SetLabel(int l){ label= l; }
	//xtĂAlmapɋL^
	void SaveIfLabeled(
		const TInNode& node, 
		TInIterator first, 
		TInIterator last, 
		TSourceLabelMaps* lmap,
		bool asAttrs= false);
	//}b`鑮̃L[B
	virtual TKey GetMatchingKey()const= 0;
	virtual void SetMatchingKey(const TKey& k)= 0;

protected:
	//p^[ɑAattrsɒǉB
	virtual void GenerateSelf(TOutAttributes& attrs, const TTargetLabelMaps* lmap)= 0;
	//xɑΉ鑮AattrsɒǉB
	virtual void GenerateLabeled(
		TOutAttributes& attrs, const TOutValues& values);

private:
	int label;//x

};
//---------------------------------------------------------------------------
//؂̃p^[
//ilƂĈ悤FIPatternImplbvĂj
template <class TTraits>
class FIPattern: public FITreeDefs<TTraits>
{
public:
	explicit FIPattern(FIPatternImpl<TTraits>* i): impl(i){}
	//m[h񂪃p^[Ƀ}b`邩𒲂ׂ
	std::pair<bool, TInIterator> Match(
		const TInNode& pnode, 
		TInIterator first, 
		TInIterator last, 
		TSourceLabelMaps* lmap)const{
			return impl->Match(pnode, first, last, lmap);
	}
	//p^[𐗌^ƂđAattrsɒǉB
	//ɃxtĂAxɑΉ鑮ԂB
	void Generate(TOutAttributes& attrs, const TTargetLabelMaps* lmap)const{
		impl->Generate(attrs, lmap);
	}
	//xt
	FIPattern& operator[](int l){
		impl->SetLabel(l);
		return *this;
	}
	//gFIOptionalPatternImplłŕԂ
	FIPattern operator!();
	//
	boost::shared_ptr<FIPatternImpl<TTraits> > GetImpl()const{ return impl; }

private:
	boost::shared_ptr<FIPatternImpl<TTraits> > impl;//

};
//---------------------------------------------------------------------------
//p^[Ȃ
template <class T>
inline FIPattern<T> operator>>(
  const FIPattern<T>& firstPattern, const FIPattern<T>& secondPattern);
//---------------------------------------------------------------------------
//PƂ̃m[h\p^[
template <class TTraits>
class FINodePatternImpl: public FIPatternImpl<TTraits>
{
public:
	//OƎqm[h̃p^[w肷
	FINodePatternImpl(const TKey& k, const FIPattern<TTraits>& p)
		: key(k), childrenPattern(p){}
	//m[h񂪃p^[Ƀ}b`邩𒲂ׂ
	virtual std::pair<bool, TInIterator>
		Match(const TInNode& pnode, TInIterator first, TInIterator last, TSourceLabelMaps* lmap);

protected:
	//p^[ɑAattrsɒǉB
	virtual void GenerateSelf(TOutAttributes& attrs, const TTargetLabelMaps* lmap);
	//}b`鑮̃L[B
	virtual TKey GetMatchingKey()const{ return key; }
	virtual void SetMatchingKey(const TKey& k);

private:
	TKey key;//}b`鑮̃L[
	FIPattern<TTraits> childrenPattern;//qm[h̃p^[

};
//---------------------------------------------------------------------------
//PƂ̕\p^[
template <class TTraits>
class FIStringPatternImpl: public FIPatternImpl<TTraits>
{
public:
	FIStringPatternImpl(const TKey& k, const TString& v)
		: key(k), str(v){}
	//m[h񂪃p^[Ƀ}b`邩𒲂ׂ
	virtual std::pair<bool, TInIterator>
		Match(const TInNode& pnode, TInIterator first, TInIterator last, TSourceLabelMaps* lmap);

protected:
	//p^[ɑAattrsɒǉB
	virtual void GenerateSelf(TOutAttributes& attrs, const TTargetLabelMaps* lmap);
	//}b`鑮̃L[B
	virtual TKey GetMatchingKey()const{ return key; }
	virtual void SetMatchingKey(const TKey& k);

private:
	TKey key;//}b`鑮̃L[
	TString str;//

};
//---------------------------------------------------------------------------
//Cӂ̒l\p^[
template <class TTraits>
class FIAnyValuePatternImpl: public FIPatternImpl<TTraits>
{
public:
	FIAnyValuePatternImpl(const TKey& k)
		: key(k){}
	//m[h񂪃p^[Ƀ}b`邩𒲂ׂ
	virtual std::pair<bool, TInIterator>
		Match(const TInNode& pnode, TInIterator first, TInIterator last, TSourceLabelMaps* lmap);

protected:
	//p^[ɑAattrsɒǉB
	virtual void GenerateSelf(TOutAttributes& attrs, const TTargetLabelMaps* lmap);
	//}b`鑮̃L[B
	virtual TKey GetMatchingKey()const{ return key; }
	virtual void SetMatchingKey(const TKey& k){ key= k; }

private:
	TKey key;//}b`鑮̃L[

};
//---------------------------------------------------------------------------
//p^[̂Ȃ
//iLqɂ >> ŕ\j
template <class TTraits>
class FIConjunctionPatternImpl: public FIPatternImpl<TTraits>
{
public:
	FIConjunctionPatternImpl(const FIPattern<TTraits>& f, const FIPattern<TTraits>& s)
		: firstPattern(f), secondPattern(s){}
	//m[h񂪃p^[Ƀ}b`邩𒲂ׂ
	virtual std::pair<bool, TInIterator>
		Match(const TInNode& pnode, TInIterator first, TInIterator last, TSourceLabelMaps* lmap);

protected:
	//p^[ɑAattrsɒǉB
	virtual void GenerateSelf(TOutAttributes& attrs, const TTargetLabelMaps* lmap);
	//}b`鑮̃L[B
	virtual TKey GetMatchingKey()const;
	virtual void SetMatchingKey(const TKey& k);

private:
	FIPattern<TTraits> firstPattern;//1ڂ̃p^[
	FIPattern<TTraits> secondPattern;//2ڂ̃p^[

};
//---------------------------------------------------------------------------
//ȗ\\p^[
//iLqɂ ! ŕ\j
template <class TTraits>
class FIOptionalPatternImpl: public FIPatternImpl<TTraits>
{
public:
	FIOptionalPatternImpl(const FIPattern<TTraits>& p)
		: pattern(p){}
	//m[h񂪃p^[Ƀ}b`邩𒲂ׂ
	virtual std::pair<bool, TInIterator>
		Match(const TInNode& pnode, TInIterator first, TInIterator last, TSourceLabelMaps* lmap);

protected:
	//p^[ɑAattrsɒǉB
	virtual void GenerateSelf(TOutAttributes& attrs, const TTargetLabelMaps* lmap);
	//}b`鑮̃L[B
	virtual TKey GetMatchingKey()const;
	virtual void SetMatchingKey(const TKey& k);

private:
	FIPattern<TTraits> pattern;//p^[

};
//---------------------------------------------------------------------------
//Cӂ̒l̑Ƀ}b`p^[
template <class TTraits>
class FIAnyPatternImpl: public FIPatternImpl<TTraits>
{
public:
	FIAnyPatternImpl(const TKey& k)
		: key(k){}
	//m[h񂪃p^[Ƀ}b`邩𒲂ׂ
	virtual std::pair<bool, TInIterator>
		Match(const TInNode& pnode, TInIterator first, TInIterator last, TSourceLabelMaps* lmap);
	//}b`鑮̃L[B
	virtual TKey GetMatchingKey()const{ return key; }
	virtual void SetMatchingKey(const TKey& k){ key= k; }

protected:
	//p^[ɑAattrsɒǉB
	virtual void GenerateSelf(TOutAttributes& attrs, const TTargetLabelMaps* lmap);

private:
	TKey key;//}b`鑮̃L[

};
//---------------------------------------------------------------------------
//Cӂ̑Ƀ}b`p^[B
//xlł͂ȂɃ}b`B
template <class TTraits>
class FIAnyAttributesPatternImpl: public FIPatternImpl<TTraits>
{
public:
	//m[h񂪃p^[Ƀ}b`邩𒲂ׂ
	virtual std::pair<bool, TInIterator>
		Match(const TInNode& pnode, TInIterator first, TInIterator last, TSourceLabelMaps* lmap);
	//}b`鑮̃L[B
	virtual TKey GetMatchingKey()const{ assert(false); return TKey(); }
	virtual void SetMatchingKey(const TKey& k){ assert(false); }

protected:
	//p^[ɑAattrsɒǉB
	virtual void GenerateSelf(TOutAttributes& attrs, const TTargetLabelMaps* lmap);

};

//---------------------------------------------------------------------------
//Cӂ̑Ƀ}b`p^[B
//xlł͂ȂɃ}b`B
template <class TTraits>
class FIAASPatternImpl: public FIPatternImpl<TTraits>
{
	FIPattern<TTraits> end;
public:
	FIAASPatternImpl(const FIPattern<TTraits>& e):end(e){
	}
	//m[h񂪃p^[Ƀ}b`邩𒲂ׂ
	virtual std::pair<bool, TInIterator>
		Match(const TInNode& pnode, TInIterator first, TInIterator last, TSourceLabelMaps* lmap);
	//}b`鑮̃L[B
	virtual TKey GetMatchingKey()const{ assert(false); return TKey(); }
	virtual void SetMatchingKey(const TKey& k){ assert(false); }

protected:
	//p^[ɑAattrsɒǉB
	virtual void GenerateSelf(TOutAttributes& attrs, const TTargetLabelMaps* lmap);
};
//---------------------------------------------------------------------------
//؂̃p^[̒`W߂
template <class TTraits>
struct FITreePatterns: public FITreeDefs<TTraits>
{
	//p^[
	typedef FIPattern<TTraits> TPattern;

	//CӃm[h
	static TPattern Any(){
		return TPattern(new FIAnyPatternImpl<TTraits>(TKey()));
	}
	//m[h
	static TPattern Node(const TPattern& childrenPattern){
		return TPattern(new FINodePatternImpl<TTraits>(TKey(), childrenPattern));
	}
	//Cӂ̒l
	static TPattern Value(){
		return TPattern(new FIAnyValuePatternImpl<TTraits>(TKey()));
	}
	//
	static TPattern String(const TString& str){
		return TPattern(new FIStringPatternImpl<TTraits>(TKey(), str));
	}
	//
	static TPattern Attribute(const TKey& key, const TPattern& pattern){
		pattern.GetImpl()->SetMatchingKey(key);
		return pattern;
	}
	//Cӑ
	static TPattern AnyAttributes(){
		return TPattern(new FIAnyAttributesPatternImpl<TTraits>());
	}
	static TPattern AAS(TPattern& p){
		return TPattern(new FIAASPatternImpl<TTraits>(p));
	}

};
//---------------------------------------------------------------------------
//ЕiTLTraitsTRTraitsj̖ؕϊ
template <class TLTraits, class TRTraits>
class FIHalfTreeConverter
{
	typedef FIHalfTreeConverter TThis;
	typedef FITreePatterns<TLTraits> TL;
	typedef FITreePatterns<TRTraits> TR;

public:
	//[ǉ
	void Add(
		const typename TL::TPattern& lpattern, 
		const typename TR::TPattern& rpattern);
	//parent[first,last)͈̔͂̑ϊǍʂoutAttrsɒǉB
	void Convert(
		typename TR::TOutAttributes& outAttrs,
		const typename TL::TInNode& parent,
		typename TL::TInIterator first, 
		typename TL::TInIterator last);

private:
	typedef std::list<std::pair<typename TL::TPattern, typename TR::TPattern> >
		TRules;

	TRules rules;//[

	//lϊB
	void ConvertValue(
		typename TR::TOutValue& outValue, 
		typename const TL::TInValue& inValue);

};
//*1 ̂TThis::tȂƁAVC7.1Ŋ֐̒`Ɛ錾vȂƌB
//---------------------------------------------------------------------------
//L[ƕ̕ϊ@̒`B
//ϊƕϊŁAL[(Traits::TKey)ƕ(Traits::TValue)̌^Ɍ݊
//ꍇ́AFIConvertationTraitsꉻāAK؂ConvertKeyConvertString
//`KvB
template <class TL, class TR>
struct FIConvertationTraits
{
	static typename TR::TKey ConvertKey(const typename TL::TKey& key){
		return key;
	}
	static typename TR::TString ConvertString(const typename TL::TString& str){
		return str;
	}
	
};
//---------------------------------------------------------------------------
//eev[gNX̃o֐̒`includeB
#include "FITreeConverterImpl.h"
//---------------------------------------------------------------------------
#endif
