#ifndef PH_JOINTMULTI_H
#define PH_JOINTMULTI_H

#include <Physics/PHJoint.h>

namespace Spr{;

///	Rx̊֐߂̊{NX
template <int NDOF>
class PHJointMulti:public PHJointBase{
public:
	typedef PTM::TVector<NDOF, double> AVec;
	AVec torque;							///<	gN
	AVec accel;								///<	֐߉x
	AVec delta_position;					///<	ψʂ̕ϕ
	AVec velocity;							///<	x
	void AddTorque(AVec t){ torque+=t; }	///<	gNǉ
	void SetTorque(AVec t){ torque=t; }		///<	gNݒ
	AVec GetTorque(){ return torque; }		///<	gN擾
	AVec GetVelocity(){ return velocity; }	///<	֐ߑx擾
	///	֐߂̎Rx
	virtual int GetJointDof(){ return NDOF; }
	///	֐ߑx̎擾
	virtual double GetJointVelocity(int i){ return velocity[i]; }
	///	֐߉x̎擾
	virtual double GetJointAccel(int i){ return accel[i]; }
	///	֐߃gN̎擾
	virtual double GetJointTorque(int i){ return torque[i]; }
	///	֐߃gN̐ݒ
	virtual void SetJointTorque(double v, int i){ torque[i] = v; }
	///	֐߃gN
	virtual void AddJointTorque(double v, int i){ torque[i] += v; }
	///	articulated inertia & ZA-force
	void CompArticulatedInertia(double dt){
		//	qWCgIa,ZǎvZD
		PHJointBase::CompArticulatedInertia(dt);
		Iss = (S_tr * Ia * S).inv();
		//em[hZa,IaɎZa,Iaςݏグ
		//Ia
		OfParent(&PHJointMulti::Ia) += 
			pXc_Mat_cXp( Ia - (Ia * S*Iss*S_tr * Ia) );
		//Za
		OfParent(&PHJointMulti::Za) += 
			pXc_Vec( Z_plus_Ia_c + Ia*S*Iss*(torque - S_tr*Z_plus_Ia_c) );
	}
	/**	hNXIntegrate() Ăяo֐D
		hNXIntegrate()̗F
		void Integrate(double dt){
			PreIntegrate(dt);
			//	ŁCdelta_position C֐߂̎pvZD
			PropagateState();
			PHJointBase::Integrate(dt);
		}
	*/
	void PreIntegrate(SGScene* scene){
		double dt = scene->GetTimeStep();
		a_p = cXp_Vec(OfParent(&PHJointMulti::a));

		if (intType == SYMPLETIC){
			//	x(dt) = x(0) + dt*v(0)/m
			//	v(dt) = v(0) + dt*f(dt)
			//xvZ
			accel = Iss*(torque - S_tr*(Z_plus_Ia_c + Ia*a_p));
			//xϕ
			velocity += accel * scene->GetTimeStep();
			//ʒuϕ鏀
			delta_position = velocity * scene->GetTimeStep();
		}else{
			//xvZ
			accel = Iss*(torque - S_tr*(Z_plus_Ia_c + Ia*a_p));
			//ʒuϕ鏀
			delta_position = (velocity + 0.5 * accel * dt) * dt;
			//xϕ
			velocity += accel * dt;
		}
		velocity *= scene->GetVelocityLossPerStep();
		//dS̉x(qm[h̐ϕŎgp)
		a = a_p + c + S*accel;
		if (velocity.norm() > PHJOINT_MAX_VELOCITY){
			DSTR << "The velocity of " << GetName() << ":" << velocity << "was limited" << std::endl;
			if (_finite(velocity[0]) ){
				velocity = PHJOINT_MAX_VELOCITY * velocity.unit();
			}else{
				velocity.clear();
			}
		}
	}
	void CalcAccel(double dt){				///<	̃WCg̉xvZDڍׂ͊{NX̃RgQƁD
		GetParent()->CalcAccel(dt);	//	ẻxvZ
		a_p = cXp_Vec(OfParent(&PHJointMulti::a));
		//xvZ
		accel = Iss*(torque - S_tr*(Z_plus_Ia_c + Ia*a_p));
		//dS̉x(qm[h̐ϕŎgp)
		a = a_p + c + S*accel;
	}
protected:
	typedef PTM::TMatrixCol<6, NDOF, double>	SMat;
	typedef PTM::TMatrixCol<NDOF, 6, double>	SMatTr;
	SMat S;									///<	spatial joint axis in Fc coord.
	SMatTr S_tr;							///<	S' S̓Ǝ̓]n.
	PTM::TMatrixCol<NDOF, NDOF, double> Iss;///<	S'IaS ̃LbV
	virtual void Reset(){
		accel.clear();
		torque.clear();
		PHJointBase::Reset();
	}
/*
	///	Ԃ̓ǂݏo
	virtual void LoadState(const SGBehaviorStates& states);
	///	Ԃ̕ۑ
	virtual void SaveState(SGBehaviorStates& states) const;
*/	
	virtual void ClearTorqueRecursive(){
		PHJointBase::ClearTorqueRecursive();
		torque.clear();
	}
};


/**	{[WCgC͎q̑ɌŒ肳D*/
class PHJointBall:public PHJointMulti<3>{
	double massFactor;
public:
	SGOBJECTDEF(PHJointBall);
	Quaterniond position;					///<	֐߂̌
	Vec3d center;			//	̒S
	double minDot;			//	͈
	double minTwist;			//	Ђ˂̉͈
	double maxTwist;			//	Ђ˂̉͈

	PHJointBall();
	///	֐߈ʒu̎擾
	virtual double GetJointPosition(int i){ return position.rotation()[i]; }

protected:
	virtual void Integrate(SGScene* scene);
	virtual void CompJointAxis();
	virtual void CompRelativePosition();
	virtual void CompRelativeVelocity();
	virtual void CompCoriolisAccel();
	virtual void Loaded(SGScene* s);
	virtual double MassFactor();
	///	Ԃ̓ǂݏo
	virtual void LoadState(const SGBehaviorStates& states);
	///	Ԃ̕ۑ
	virtual void SaveState(SGBehaviorStates& states) const;
};

/**	jo[TWCgC͎q̑ɌŒ肳 x yD*/
class PHJointUniversal:public PHJointMulti<2>{
public:
	SGOBJECTDEF(PHJointUniversal);
	Vec2d position;					///<	֐߂̌
	///	֐߈ʒu̎擾
	virtual double GetJointPosition(int i){ return position[i]; }

	virtual void Integrate(SGScene* scene);
	virtual void CompJointAxis();
	virtual void CompRelativePosition();
	virtual void CompRelativeVelocity();
	virtual void CompCoriolisAccel();
	///	Ԃ̓ǂݏo
	virtual void LoadState(const SGBehaviorStates& states);
	///	Ԃ̕ۑ
	virtual void SaveState(SGBehaviorStates& states) const;
};

}
#endif
